import logging
from typing import Any, Iterable

import networkx as nx
import numpy as np
from scipy.spatial import KDTree
from skimage.measure import regionprops
from tqdm import tqdm

from .graph_attributes import NodeAttr

logger = logging.getLogger(__name__)


def get_node_id(time: int, label_id: int, hypothesis_id: int | None = None) -> str:
    """Construct a node id given the time frame, segmentation label id, and
    optionally the hypothesis id. This function is not designed for candidate graphs
    that do not come from segmentations, but could be used if there is a similar
    "detection id" that is unique for all cells detected in a given frame.

    Args:
        time (int): The time frame the node is in
        label_id (int): The label the node has in the segmentation.
        hypothesis_id (int | None, optional): An integer representing which hypothesis
            the segmentation came from, if applicable. Defaults to None.

    Returns:
        str: A string to use as the node id in the candidate graph. Assuming that label
        ids are not repeated in the same time frame and hypothesis, it is unique.
    """
    if hypothesis_id is not None:
        return f"{time}_{hypothesis_id}_{label_id}"
    else:
        return f"{time}_{label_id}"


def nodes_from_segmentation(
    segmentation: np.ndarray,
) -> tuple[nx.DiGraph, dict[int, list[Any]]]:
    """Extract candidate nodes from a segmentation. Also computes specified attributes.
    Returns a networkx graph with only nodes, and also a dictionary from frames to
    node_ids for efficient edge adding.

    Args:
        segmentation (np.ndarray): A numpy array with integer labels and dimensions
            (t, h, [z], y, x), where h is the number of hypotheses.

    Returns:
        tuple[nx.DiGraph, dict[int, list[Any]]]: A candidate graph with only nodes,
            and a mapping from time frames to node ids.
    """
    cand_graph = nx.DiGraph()
    # also construct a dictionary from time frame to node_id for efficiency
    node_frame_dict: dict[int, list[Any]] = {}
    print("Extracting nodes from segmentation")
    num_hypotheses = segmentation.shape[1]
    for t in tqdm(range(len(segmentation))):
        segs = segmentation[t]
        hypo_id: int | None
        for hypo_id, hypo in enumerate(segs):
            if num_hypotheses == 1:
                hypo_id = None
            nodes_in_frame = []
            props = regionprops(hypo)
            for regionprop in props:
                node_id = get_node_id(t, regionprop.label, hypothesis_id=hypo_id)
                attrs = {
                    NodeAttr.TIME.value: t,
                }
                attrs[NodeAttr.SEG_ID.value] = regionprop.label
                if hypo_id is not None:
                    attrs[NodeAttr.SEG_HYPO.value] = hypo_id
                centroid = regionprop.centroid  # [z,] y, x
                attrs[NodeAttr.POS.value] = centroid
                cand_graph.add_node(node_id, **attrs)
                nodes_in_frame.append(node_id)
            if nodes_in_frame:
                if t not in node_frame_dict:
                    node_frame_dict[t] = []
                node_frame_dict[t].extend(nodes_in_frame)
    return cand_graph, node_frame_dict


def nodes_from_points_list(
    points_list: np.ndarray,
) -> tuple[nx.DiGraph, dict[int, list[Any]]]:
    """Extract candidate nodes from a list of points. Uses the index of the
    point in the list as its unique id.
    Returns a networkx graph with only nodes, and also a dictionary from frames to
    node_ids for efficient edge adding.

    Args:
        points_list (np.ndarray): An NxD numpy array with N points and D
            (3 or 4) dimensions. Dimensions should be in order (t, [z], y, x).

    Returns:
        tuple[nx.DiGraph, dict[int, list[Any]]]: A candidate graph with only nodes,
            and a mapping from time frames to node ids.
    """
    cand_graph = nx.DiGraph()
    # also construct a dictionary from time frame to node_id for efficiency
    node_frame_dict: dict[int, list[Any]] = {}
    print("Extracting nodes from points list")
    for i, point in enumerate(points_list):
        # assume t, [z], y, x
        t = point[0]
        pos = point[1:]
        node_id = i
        attrs = {
            NodeAttr.TIME.value: t,
            NodeAttr.POS.value: pos,
        }
        cand_graph.add_node(node_id, **attrs)
        if t not in node_frame_dict:
            node_frame_dict[t] = []
        node_frame_dict[t].append(node_id)
    return cand_graph, node_frame_dict


def _compute_node_frame_dict(cand_graph: nx.DiGraph) -> dict[int, list[Any]]:
    """Compute dictionary from time frames to node ids for candidate graph.

    Args:
        cand_graph (nx.DiGraph): A networkx graph

    Returns:
        dict[int, list[Any]]: A mapping from time frames to lists of node ids.
    """
    node_frame_dict: dict[int, list[Any]] = {}
    for node, data in cand_graph.nodes(data=True):
        t = data[NodeAttr.TIME.value]
        if t not in node_frame_dict:
            node_frame_dict[t] = []
        node_frame_dict[t].append(node)
    return node_frame_dict


def create_kdtree(cand_graph: nx.DiGraph, node_ids: Iterable[Any]) -> KDTree:
    positions = [cand_graph.nodes[node][NodeAttr.POS.value] for node in node_ids]
    return KDTree(positions)


def add_cand_edges(
    cand_graph: nx.DiGraph,
    max_edge_distance: float,
    node_frame_dict: None | dict[int, list[Any]] = None,
) -> None:
    """Add candidate edges to a candidate graph by connecting all nodes in adjacent
    frames that are closer than max_edge_distance. Also adds attributes to the edges.

    Args:
        cand_graph (nx.DiGraph): Candidate graph with only nodes populated. Will
            be modified in-place to add edges.
        max_edge_distance (float): Maximum distance that objects can travel between
            frames. All nodes within this distance in adjacent frames will by connected
            with a candidate edge.
        node_frame_dict (dict[int, list[Any]] | None, optional): A mapping from frames
            to node ids. If not provided, it will be computed from cand_graph. Defaults
            to None.
    """
    print("Extracting candidate edges")
    if not node_frame_dict:
        node_frame_dict = _compute_node_frame_dict(cand_graph)

    frames = sorted(node_frame_dict.keys())
    prev_node_ids = node_frame_dict[frames[0]]
    prev_kdtree = create_kdtree(cand_graph, prev_node_ids)
    for frame in tqdm(frames):
        if frame + 1 not in node_frame_dict:
            continue
        next_node_ids = node_frame_dict[frame + 1]
        next_kdtree = create_kdtree(cand_graph, next_node_ids)

        matched_indices = prev_kdtree.query_ball_tree(next_kdtree, max_edge_distance)

        for prev_node_id, next_node_indices in zip(prev_node_ids, matched_indices):
            for next_node_index in next_node_indices:
                next_node_id = next_node_ids[next_node_index]
                cand_graph.add_edge(prev_node_id, next_node_id)

        prev_node_ids = next_node_ids
        prev_kdtree = next_kdtree



import networkx as nx
import numpy as np

from motile_toolbox.candidate_graph import NodeAttr


def assign_tracklet_ids(graph: nx.DiGraph) -> nx.DiGraph:
    """Add a tracklet_id attribute to a graph by removing division edges,
    assigning one id to each connected component.
    Designed as a helper for visualizing the graph in the napari Tracks layer.

    Args:
        graph (nx.DiGraph): A networkx graph with a tracking solution

    Returns:
        nx.DiGraph: The same graph with the tracklet_id assigned. Probably
        occurrs in place but returned just to be clear.
    """
    graph_copy = graph.copy()

    parents = [node for node, degree in graph.out_degree() if degree >= 2]
    intertrack_edges = []

    # Remove all intertrack edges from a copy of the original graph
    for parent in parents:
        daughters = [child for p, child in graph.out_edges(parent)]
        for daughter in daughters:
            graph_copy.remove_edge(parent, daughter)
            intertrack_edges.append((parent, daughter))

    track_id = 0
    for tracklet in nx.weakly_connected_components(graph_copy):
        nx.set_node_attributes(
            graph, {node: {"tracklet_id": track_id} for node in tracklet}
        )
        track_id += 1
    return graph, intertrack_edges


def to_napari_tracks_layer(
    graph, frame_key=NodeAttr.TIME.value, location_key=NodeAttr.POS.value, properties=()
):
    """Function to take a networkx graph and return the data needed to add to
    a napari tracks layer.

    Args:
        graph (nx.DiGraph): _description_
        frame_key (str, optional): Key in graph attributes containing time frame.
            Defaults to NodeAttr.TIME.value.
        location_key (str, optional): Key in graph node attributes containing
            location. Defaults to NodeAttr.POS.value.
        properties (tuple, optional): Keys in graph node attributes to add
            to the visualization layer. Defaults to (). NOTE: not working now :(

    Returns:
        data : array (N, D+1)
            Coordinates for N points in D+1 dimensions. ID,T,(Z),Y,X. The first
            axis is the integer ID of the track. D is either 3 or 4 for planar
            or volumetric timeseries respectively.
        properties : dict {str: array (N,)}
            Properties for each point. Each property should be an array of length N,
            where N is the number of points.
        graph : dict {int: list}
            Graph representing associations between tracks. Dictionary defines the
            mapping between a track ID and the parents of the track. This can be
            one (the track has one parent, and the parent has >=1 child) in the
            case of track splitting, or more than one (the track has multiple
            parents, but only one child) in the case of track merging.
    """
    for _, loc in graph.nodes(data=location_key):
        ndim = len(loc)
        break
    napari_data = np.zeros((graph.number_of_nodes(), ndim + 2))
    napari_properties = {prop: np.zeros(graph.number_of_nodes()) for prop in properties}
    napari_edges = {}
    graph, intertrack_edges = assign_tracklet_ids(graph)
    for index, node in enumerate(graph.nodes(data=True)):
        node_id, data = node
        location = data[location_key]
        napari_data[index] = [data["tracklet_id"], data[frame_key], *location]
        for prop in properties:
            if prop in data:
                napari_properties[prop][index] = data[prop]
    napari_edges = {}
    for parent, child in intertrack_edges:
        parent_track_id = graph.nodes[parent]["tracklet_id"]
        child_track_id = graph.nodes[child]["tracklet_id"]
        if child_track_id in napari_edges:
            napari_edges[child_track_id].append(parent_track_id)
        else:
            napari_edges[child_track_id] = [parent_track_id]
    return napari_data, napari_properties, napari_edges



import networkx as nx
import pytest
from motile_toolbox.candidate_graph import EdgeAttr, add_iou
from motile_toolbox.candidate_graph.iou import _compute_ious


def test_compute_ious_2d(segmentation_2d):
    ious = _compute_ious(segmentation_2d[0], segmentation_2d[1])
    expected = [
        (1, 2, 555.46 / 1408.0),
    ]
    for iou, expected_iou in zip(ious, expected):
        assert iou == pytest.approx(expected_iou, abs=0.01)

    ious = _compute_ious(segmentation_2d[1], segmentation_2d[1])
    expected = [(1, 1, 1.0), (2, 2, 1.0)]
    for iou, expected_iou in zip(ious, expected):
        assert iou == pytest.approx(expected_iou, abs=0.01)


def test_compute_ious_3d(segmentation_3d):
    ious = _compute_ious(segmentation_3d[0], segmentation_3d[1])
    expected = [(1, 2, 0.30)]
    for iou, expected_iou in zip(ious, expected):
        assert iou == pytest.approx(expected_iou, abs=0.01)

    ious = _compute_ious(segmentation_3d[1], segmentation_3d[1])
    expected = [(1, 1, 1.0), (2, 2, 1.0)]
    for iou, expected_iou in zip(ious, expected):
        assert iou == pytest.approx(expected_iou, abs=0.01)


def test_add_iou_2d(segmentation_2d, graph_2d):
    expected = graph_2d
    input_graph = graph_2d.copy()
    nx.set_edge_attributes(input_graph, -1, name=EdgeAttr.IOU.value)
    add_iou(input_graph, segmentation_2d)
    for s, t, attrs in expected.edges(data=True):
        assert (
            pytest.approx(attrs[EdgeAttr.IOU.value], abs=0.01)
            == input_graph.edges[(s, t)][EdgeAttr.IOU.value]
        )


def test_multi_hypo_iou_2d(multi_hypothesis_segmentation_2d, multi_hypothesis_graph_2d):
    expected = multi_hypothesis_graph_2d
    input_graph = multi_hypothesis_graph_2d.copy()
    nx.set_edge_attributes(input_graph, -1, name=EdgeAttr.IOU.value)
    add_iou(input_graph, multi_hypothesis_segmentation_2d)
    for s, t, attrs in expected.edges(data=True):
        print(s, t)
        assert (
            pytest.approx(attrs[EdgeAttr.IOU.value], abs=0.01)
            == input_graph.edges[(s, t)][EdgeAttr.IOU.value]
        )



import numpy as np
from motile_toolbox.utils import relabel_segmentation
from numpy.testing import assert_array_equal
from skimage.draw import disk


def test_relabel_segmentation(segmentation_2d, graph_2d):
    frame_shape = (100, 100)
    expected = np.zeros(segmentation_2d.shape, dtype="int32")
    # make frame with one cell in center with label 1
    rr, cc = disk(center=(50, 50), radius=20, shape=(100, 100))
    expected[0, 0][rr, cc] = 1

    # make frame with cell centered at (20, 80) with label 1
    rr, cc = disk(center=(20, 80), radius=10, shape=frame_shape)
    expected[1, 0][rr, cc] = 1

    graph_2d.remove_node("1_2")
    relabeled_seg = relabel_segmentation(graph_2d, segmentation_2d)
    print(f"Nonzero relabeled: {np.count_nonzero(relabeled_seg)}")
    print(f"Nonzero expected: {np.count_nonzero(expected)}")
    print(f"Max relabeled: {np.max(relabeled_seg)}")
    print(f"Max expected: {np.max(expected)}")

    assert_array_equal(relabeled_seg, expected)



import logging
from typing import Any

import networkx as nx
import numpy as np

from .conflict_sets import compute_conflict_sets
from .iou import add_iou
from .utils import add_cand_edges, nodes_from_points_list, nodes_from_segmentation

logger = logging.getLogger(__name__)


def get_candidate_graph(
    segmentation: np.ndarray,
    max_edge_distance: float,
    iou: bool = False,
) -> tuple[nx.DiGraph, list[set[Any]] | None]:
    """Construct a candidate graph from a segmentation array. Nodes are placed at the
    centroid of each segmentation and edges are added for all nodes in adjacent frames
    within max_edge_distance. If segmentation contains multiple hypotheses, will also
    return a list of conflicting node ids that cannot be selected together.

    Args:
        segmentation (np.ndarray): A numpy array with integer labels and dimensions
            (t, h, [z], y, x), where h is the number of hypotheses.
        max_edge_distance (float): Maximum distance that objects can travel between
            frames. All nodes with centroids within this distance in adjacent frames
            will by connected with a candidate edge.
        iou (bool, optional): Whether to include IOU on the candidate graph.
            Defaults to False.

    Returns:
        tuple[nx.DiGraph, list[set[Any]] | None]: A candidate graph that can be passed
        to the motile solver, and a list of conflicting node ids.
    """
    num_hypotheses = segmentation.shape[1]

    # add nodes
    cand_graph, node_frame_dict = nodes_from_segmentation(segmentation)
    logger.info(f"Candidate nodes: {cand_graph.number_of_nodes()}")

    # add edges
    add_cand_edges(
        cand_graph,
        max_edge_distance=max_edge_distance,
        node_frame_dict=node_frame_dict,
    )
    if iou:
        add_iou(cand_graph, segmentation, node_frame_dict)

    logger.info(f"Candidate edges: {cand_graph.number_of_edges()}")

    # Compute conflict sets between segmentations
    conflicts = []
    if num_hypotheses > 1:
        for time, segs in enumerate(segmentation):
            conflicts.extend(compute_conflict_sets(segs, time))

    return cand_graph, conflicts


def get_candidate_graph_from_points_list(
    points_list: np.ndarray,
    max_edge_distance: float,
) -> nx.DiGraph:
    """Construct a candidate graph from a points list.

    Args:
        points_list (np.ndarray): An NxD numpy array with N points and D
            (3 or 4) dimensions. Dimensions should be in order  (t, [z], y, x).
        max_edge_distance (float): Maximum distance that objects can travel between
            frames. All nodes with centroids within this distance in adjacent frames
            will by connected with a candidate edge.

    Returns:
        nx.DiGraph: A candidate graph that can be passed to the motile solver.
        Multiple hypotheses not supported for points input.
    """
    # add nodes
    cand_graph, node_frame_dict = nodes_from_points_list(points_list)
    logger.info(f"Candidate nodes: {cand_graph.number_of_nodes()}")
    # add edges
    add_cand_edges(
        cand_graph,
        max_edge_distance=max_edge_distance,
        node_frame_dict=node_frame_dict,
    )
    return cand_graph



import csv
import logging
from typing import Any, MutableMapping

import networkx as nx
import numpy as np
import zarr

logger = logging.getLogger(__name__)


def load_csv_tracks(
    tracks_path: str, frames: tuple[int, int] | None = None
) -> nx.DiGraph:
    """Load tracks from a csv to a networkx graph.
    Expects the following tab-separated columns (from mskcc-confocal):
        t z y x cell_id parent_id track_id radius name div_state
    But everything after parent_id is optional, as technically are z y and x.

    Args:
        tracks_path (str): path to tracks file
        frames (tuple, optional): Tuple of start frame, end frame to limit the tracks to
            these time points. Includes start frame, excludes end frame. Defaults to
            None.

    Returns: nx.DiGraph where nodes have all the attributes from the columns of the csv,
        and edges have no attributes. Edges go forward in time from parent to child.
    """
    graph = nx.DiGraph()
    with open(tracks_path) as f:
        reader = csv.DictReader(f, delimiter="\t")
        # t	z	y	x	cell_id	parent_id	track_id	radius	name	div_state
        for row in reader:
            cell = _convert_types(row)
            if frames:
                time = cell["t"]
                if time < frames[0] or time >= frames[1]:
                    continue
            cell_id = cell["cell_id"]
            graph.add_node(cell["cell_id"], **cell)
            parent_id = cell["parent_id"]
            if parent_id != -1:
                # only add edge if it falls within the specified frames
                if (not frames) or time > frames[0]:
                    graph.add_edge(parent_id, cell_id)
    logger.info(
        f"Loaded {graph.number_of_nodes()} nodes and {graph.number_of_edges()} edges "
        f"from {tracks_path}."
    )
    return graph


def _convert_types(row: dict[str, str]) -> MutableMapping[str, Any]:
    """Helper function for loading the tracks csv with correct types
    Converts certain columns to int and others to float - unlisted
    columns are left as strings.

    Args:
        row (dict[str, str]): Row from csv.DictReader

    Returns:
        dict: Same row with the types converted from strings to ints/floats
        for the appropriate keys.
    """
    int_vals = ["t", "cell_id", "parent_id", "track_id", "div_state"]
    float_vals = ["z", "y", "x", "radius"]
    converted_row: MutableMapping[str, Any] = {}
    for key, val in row.items():
        if key in int_vals:
            converted_row[key] = int(val)
        elif key in float_vals:
            converted_row[key] = float(val)
        else:
            converted_row[key] = val
    return converted_row


def load_cellulus_results(
    path_to_zarr: str,
    image_group: str = "test/raw",
    seg_group: str = "post-processed-segmentation",
    seg_channel: int = 0,
) -> tuple[np.ndarray, np.ndarray]:
    """Load images and segmentations from custom cellulus zarr.

    Args:
        path_to_zarr (str): Path to zarr containing cellulus results.
        image_group (str, optional): Zarr group containing raw images. Defaults to
            "test/raw".
        seg_group (str, optional): Zarr group containing segmentation.
            Defaults to "post-processed-segmentation".
        seg_channel (int, optional): Channel of segmentation to use. Segmentation
            channels correspond to levels in multi-hypothesis cellulus. Defaults to 0.

    Returns:
        tuple[np.ndarray, np.ndarray]: an array with t, z, y, x for raw images and
            segmentation results.
    """
    base = zarr.open(path_to_zarr, "r")
    images = base[image_group]
    segmentation = base[seg_group][
        :, seg_channel
    ]  # orginally t, c, z, y, x. want to select channel

    # should return (t, z, y, x) for both
    return np.squeeze(images), np.squeeze(segmentation)



from .compute_graph import get_candidate_graph, get_candidate_graph_from_points_list
from .graph_attributes import EdgeAttr, NodeAttr
from .graph_to_nx import graph_to_nx
from .iou import add_iou
from .utils import add_cand_edges, get_node_id, nodes_from_segmentation



from itertools import product
from typing import Any

import networkx as nx
import numpy as np
from tqdm import tqdm

from .graph_attributes import EdgeAttr
from .utils import _compute_node_frame_dict, get_node_id


def _compute_ious(
    frame1: np.ndarray, frame2: np.ndarray
) -> list[tuple[int, int, float]]:
    """Compute label IOUs between two label arrays of the same shape. Ignores background
    (label 0).

    Args:
        frame1 (np.ndarray): Array with integer labels
        frame2 (np.ndarray): Array with integer labels

    Returns:
        list[tuple[int, int, float]]: List of tuples of label in frame 1, label in
            frame 2, and iou values. Labels that have no overlap are not included.
    """
    frame1 = frame1.flatten()
    frame2 = frame2.flatten()
    # get indices where both are not zero (ignore background)
    # this speeds up computation significantly
    non_zero_indices = np.logical_and(frame1, frame2)
    flattened_stacked = np.array([frame1[non_zero_indices], frame2[non_zero_indices]])

    values, counts = np.unique(flattened_stacked, axis=1, return_counts=True)
    frame1_values, frame1_counts = np.unique(frame1, return_counts=True)
    frame1_label_sizes = dict(zip(frame1_values, frame1_counts))
    frame2_values, frame2_counts = np.unique(frame2, return_counts=True)
    frame2_label_sizes = dict(zip(frame2_values, frame2_counts))
    ious: list[tuple[int, int, float]] = []
    for index in range(values.shape[1]):
        pair = values[:, index]
        intersection = counts[index]
        id1, id2 = pair
        union = frame1_label_sizes[id1] + frame2_label_sizes[id2] - intersection
        ious.append((id1, id2, intersection / union))
    return ious


def _get_iou_dict(segmentation) -> dict[str, dict[str, float]]:
    """Get all ious values for the provided segmentation (all frames).
    Will return as map from node_id -> dict[node_id] -> iou for easy
    navigation when adding to candidate graph.

    Args:
        segmentation (np.ndarray): Segmentation that was used to create cand_graph.
            Has shape (t, h, [z], y, x), where h is the number of hypotheses.

    Returns:
        dict[str, dict[str, float]]: A map from node id to another dictionary, which
            contains node_ids to iou values.
    """
    iou_dict: dict[str, dict[str, float]] = {}
    hypo_pairs: list[tuple[int | None, ...]]
    num_hypotheses = segmentation.shape[1]
    if num_hypotheses > 1:
        hypo_pairs = list(product(range(num_hypotheses), repeat=2))
    else:
        hypo_pairs = [(None, None)]

    for frame in range(len(segmentation) - 1):
        for hypo1, hypo2 in hypo_pairs:
            seg1 = segmentation[frame][hypo1]
            seg2 = segmentation[frame + 1][hypo2]
            ious = _compute_ious(seg1, seg2)
            for label1, label2, iou in ious:
                node_id1 = get_node_id(frame, label1, hypo1)
                if node_id1 not in iou_dict:
                    iou_dict[node_id1] = {}
                node_id2 = get_node_id(frame + 1, label2, hypo2)
                iou_dict[node_id1][node_id2] = iou
    return iou_dict


def add_iou(
    cand_graph: nx.DiGraph,
    segmentation: np.ndarray,
    node_frame_dict: dict[int, list[Any]] | None = None,
) -> None:
    """Add IOU to the candidate graph.

    Args:
        cand_graph (nx.DiGraph): Candidate graph with nodes and edges already populated
        segmentation (np.ndarray): segmentation that was used to create cand_graph.
            Has shape (t, h, [z], y, x), where h is the number of hypotheses.
        node_frame_dict(dict[int, list[Any]] | None, optional): A mapping from
            time frames to nodes in that frame. Will be computed if not provided,
            but can be provided for efficiency (e.g. after running
            nodes_from_segmentation). Defaults to None.
    """
    if node_frame_dict is None:
        node_frame_dict = _compute_node_frame_dict(cand_graph)
    frames = sorted(node_frame_dict.keys())
    ious = _get_iou_dict(segmentation)
    for frame in tqdm(frames):
        if frame + 1 not in node_frame_dict.keys():
            continue
        next_nodes = node_frame_dict[frame + 1]
        for node_id in node_frame_dict[frame]:
            for next_id in next_nodes:
                iou = ious.get(node_id, {}).get(next_id, 0)
                if (node_id, next_id) in cand_graph.edges:
                    cand_graph.edges[(node_id, next_id)][EdgeAttr.IOU.value] = iou



from collections import Counter

import networkx as nx
import numpy as np
from motile_toolbox.candidate_graph import (
    NodeAttr,
    add_cand_edges,
    get_node_id,
    nodes_from_segmentation,
)
from motile_toolbox.candidate_graph.utils import (
    _compute_node_frame_dict,
    nodes_from_points_list,
)


# nodes_from_segmentation
def test_nodes_from_segmentation_empty():
    # test with empty segmentation
    empty_graph, node_frame_dict = nodes_from_segmentation(
        np.zeros((3, 1, 10, 10), dtype="int32")
    )
    assert Counter(empty_graph.nodes) == Counter([])
    assert node_frame_dict == {}


def test_nodes_from_segmentation_2d(segmentation_2d):
    # test with 2D segmentation
    node_graph, node_frame_dict = nodes_from_segmentation(
        segmentation=segmentation_2d,
    )
    assert Counter(list(node_graph.nodes)) == Counter(["0_1", "1_1", "1_2"])
    assert node_graph.nodes["1_1"][NodeAttr.SEG_ID.value] == 1
    assert node_graph.nodes["1_1"][NodeAttr.TIME.value] == 1
    assert node_graph.nodes["1_1"][NodeAttr.POS.value] == (20, 80)

    assert node_frame_dict[0] == ["0_1"]
    assert Counter(node_frame_dict[1]) == Counter(["1_1", "1_2"])


def test_nodes_from_segmentation_2d_hypo(
    multi_hypothesis_segmentation_2d, multi_hypothesis_graph_2d
):
    # test with 2D segmentation
    node_graph, node_frame_dict = nodes_from_segmentation(
        segmentation=multi_hypothesis_segmentation_2d
    )
    assert Counter(list(node_graph.nodes)) == Counter(
        list(multi_hypothesis_graph_2d.nodes)
    )
    assert node_graph.nodes["1_0_1"][NodeAttr.SEG_ID.value] == 1
    assert node_graph.nodes["1_0_1"][NodeAttr.SEG_HYPO.value] == 0
    assert node_graph.nodes["1_0_1"][NodeAttr.TIME.value] == 1
    assert node_graph.nodes["1_0_1"][NodeAttr.POS.value] == (20, 80)

    assert Counter(node_frame_dict[0]) == Counter(["0_0_1", "0_1_1"])
    assert Counter(node_frame_dict[1]) == Counter(["1_0_1", "1_0_2", "1_1_1", "1_1_2"])


def test_nodes_from_segmentation_3d(segmentation_3d):
    # test with 3D segmentation
    node_graph, node_frame_dict = nodes_from_segmentation(
        segmentation=segmentation_3d,
    )
    assert Counter(list(node_graph.nodes)) == Counter(["0_1", "1_1", "1_2"])
    assert node_graph.nodes["1_1"][NodeAttr.SEG_ID.value] == 1
    assert node_graph.nodes["1_1"][NodeAttr.TIME.value] == 1
    assert node_graph.nodes["1_1"][NodeAttr.POS.value] == (20, 50, 80)

    assert node_frame_dict[0] == ["0_1"]
    assert Counter(node_frame_dict[1]) == Counter(["1_1", "1_2"])


# add_cand_edges
def test_add_cand_edges_2d(graph_2d):
    cand_graph = nx.create_empty_copy(graph_2d)
    add_cand_edges(cand_graph, max_edge_distance=50)
    assert Counter(list(cand_graph.edges)) == Counter(list(graph_2d.edges))


def test_add_cand_edges_3d(graph_3d):
    cand_graph = nx.create_empty_copy(graph_3d)
    add_cand_edges(cand_graph, max_edge_distance=15)
    graph_3d.remove_edge("0_1", "1_1")
    assert Counter(list(cand_graph.edges)) == Counter(list(graph_3d.edges))


def test_get_node_id():
    assert get_node_id(0, 2) == "0_2"
    assert get_node_id(2, 10, 3) == "2_3_10"


def test_compute_node_frame_dict(graph_2d):
    node_frame_dict = _compute_node_frame_dict(graph_2d)
    expected = {
        0: [
            "0_1",
        ],
        1: ["1_1", "1_2"],
    }
    assert node_frame_dict == expected


def test_nodes_from_points_list_2d():
    points_list = np.array(
        [
            [0, 1, 2, 3],
            [2, 3, 4, 5],
            [1, 2, 3, 4],
        ]
    )
    cand_graph, node_frame_dict = nodes_from_points_list(points_list)
    assert Counter(list(cand_graph.nodes)) == Counter([0, 1, 2])
    assert cand_graph.nodes[0][NodeAttr.TIME.value] == 0
    assert (cand_graph.nodes[0][NodeAttr.POS.value] == np.array([1, 2, 3])).all()
    assert cand_graph.nodes[1][NodeAttr.TIME.value] == 2
    assert (cand_graph.nodes[1][NodeAttr.POS.value] == np.array([3, 4, 5])).all()



"""
.. include:: ../../README.md
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("motile-toolbox")
except PackageNotFoundError:
    # package is not installed
    __version__ = "uninstalled"



import networkx as nx
import numpy as np

from motile_toolbox.candidate_graph import NodeAttr


def relabel_segmentation(
    solution_nx_graph: nx.DiGraph,
    segmentation: np.ndarray,
) -> np.ndarray:
    """Relabel a segmentation based on tracking results so that nodes in same
    track share the same id. IDs do change at division.

    Args:
        solution_nx_graph (nx.DiGraph): Networkx graph with the solution to use
            for relabeling. Nodes not in graph will be removed from seg. Original
            segmentation ids and hypothesis ids have to be stored in the graph so we
            can map them back.
        segmentation (np.ndarray): Original (potentially multi-hypothesis)
            segmentation with dimensions (t,h,[z],y,x), where h is 1 for single
            input segmentation.

    Returns:
        np.ndarray: Relabeled segmentation array where nodes in same track share same
            id with shape (t,1,[z],y,x)
    """
    output_shape = (segmentation.shape[0], 1, *segmentation.shape[2:])
    tracked_masks = np.zeros_like(segmentation, shape=output_shape)
    id_counter = 1
    parent_nodes = [n for (n, d) in solution_nx_graph.out_degree() if d > 1]
    soln_copy = solution_nx_graph.copy()
    for parent_node in parent_nodes:
        out_edges = solution_nx_graph.out_edges(parent_node)
        soln_copy.remove_edges_from(out_edges)
    for node_set in nx.weakly_connected_components(soln_copy):
        for node in node_set:
            time_frame = solution_nx_graph.nodes[node][NodeAttr.TIME.value]
            previous_seg_id = solution_nx_graph.nodes[node][NodeAttr.SEG_ID.value]
            if NodeAttr.SEG_HYPO.value in solution_nx_graph.nodes[node]:
                hypothesis_id = solution_nx_graph.nodes[node][NodeAttr.SEG_HYPO.value]
            else:
                hypothesis_id = 0
            previous_seg_mask = (
                segmentation[time_frame, hypothesis_id] == previous_seg_id
            )
            tracked_masks[time_frame, 0][previous_seg_mask] = id_counter
        id_counter += 1
    return tracked_masks



import networkx as nx
from motile import TrackGraph
from motile_toolbox.candidate_graph import graph_to_nx
from networkx.utils import graphs_equal


def test_graph_to_nx(graph_3d: nx.DiGraph):
    track_graph = TrackGraph(nx_graph=graph_3d, frame_attribute="time")
    nx_graph = graph_to_nx(track_graph)
    assert graphs_equal(graph_3d, nx_graph)



from collections import Counter
from pathlib import Path

from motile_toolbox.utils import loading_utils


def test_load_csv_tracks():
    tests_path = Path(__file__).parent.parent
    tracks = loading_utils.load_csv_tracks(tests_path / "data/sample_tracks.csv")
    assert tracks.number_of_nodes() == 107
    assert tracks.number_of_edges() == 105
    cell_42 = tracks.nodes[42]
    assert cell_42["t"] == 11
    assert cell_42["z"] == 100.0
    assert cell_42["y"] == 208.0
    assert cell_42["x"] == 204.0
    assert cell_42["cell_id"] == 42
    assert cell_42["parent_id"] == 38
    assert cell_42["track_id"] == 0
    assert cell_42["radius"] == 18.5
    assert cell_42["name"] == "ABa"
    assert cell_42["div_state"] == 1
    assert Counter(list(tracks.in_edges(42))) == Counter([(38, 42)])
    assert Counter(list(tracks.out_edges(42))) == Counter([(42, 46), (42, 51)])

    # limit frames
    tracks = loading_utils.load_csv_tracks(
        tests_path / "data/sample_tracks.csv", frames=(11, 13)
    )
    assert tracks.number_of_nodes() == 10
    assert tracks.number_of_edges() == 6
    assert Counter(list(tracks.in_edges(42))) == Counter([])
    assert Counter(list(tracks.out_edges(42))) == Counter([(42, 46), (42, 51)])



from itertools import combinations

import numpy as np

from .utils import (
    get_node_id,
)


def compute_conflict_sets(segmentation_frame: np.ndarray, time: int) -> list[set]:
    """Compute all sets of node ids that conflict with each other.
    Note: Results might include redundant sets, for example {a, b, c} and {a, b}
    might both appear in the results.

    Args:
        segmentation_frame (np.ndarray):  One frame of the multiple hypothesis
            segmentation. Dimensions are (h, [z], y, x), where h is the number of
            hypotheses.
        time (int): Time frame, for computing node_ids.

    Returns:
        list[set]: list of sets of node ids that overlap. Might include some sets
            that are subsets of others.
    """
    flattened_segs = [seg.flatten() for seg in segmentation_frame]

    # get locations where at least two hypotheses have labels
    # This approach may be inefficient, but likely doesn't matter compared to np.unique
    conflict_indices = np.zeros(flattened_segs[0].shape, dtype=bool)
    for seg1, seg2 in combinations(flattened_segs, 2):
        non_zero_indices = np.logical_and(seg1, seg2)
        conflict_indices = np.logical_or(conflict_indices, non_zero_indices)

    flattened_stacked = np.array([seg[conflict_indices] for seg in flattened_segs])
    values = np.unique(flattened_stacked, axis=1)
    values = np.transpose(values)
    conflict_sets = []
    for conflicting_labels in values:
        id_set = set()
        for hypo_id, label in enumerate(conflicting_labels):
            if label != 0:
                id_set.add(get_node_id(time, label, hypo_id))
        conflict_sets.append(id_set)
    return conflict_sets



import numpy as np
from motile_toolbox.candidate_graph.conflict_sets import compute_conflict_sets
from pytest_unordered import unordered


def test_conflict_sets_2d(multi_hypothesis_segmentation_2d):
    for t in range(multi_hypothesis_segmentation_2d.shape[0]):
        conflict_set = compute_conflict_sets(multi_hypothesis_segmentation_2d[t], t)
        if t == 0:
            expected = [{"0_1_1", "0_0_1"}]
            assert len(conflict_set) == 1
            assert conflict_set == unordered(expected)
        elif t == 1:
            expected = [{"1_0_2", "1_1_2"}, {"1_0_1", "1_1_1"}]
            assert len(conflict_set) == 2
            assert conflict_set == unordered(expected)


def test_conflict_sets_2d_reshaped(multi_hypothesis_segmentation_2d):
    """Reshape segmentation array just to provide a slightly difficult example."""

    reshaped = np.asarray(
        [
            multi_hypothesis_segmentation_2d[0, 0],  # hypothesis 0
            multi_hypothesis_segmentation_2d[1, 0],  # hypothesis 1
            multi_hypothesis_segmentation_2d[1, 1],
        ]
    )  # hypothesis 2
    conflict_set = compute_conflict_sets(reshaped, 0)
    # note the expected ids are not really there since the
    # reshaped array is artifically constructed
    expected = [
        {"0_0_1", "0_1_2", "0_2_2"},
        {"0_1_1", "0_2_1"},
        {"0_0_1", "0_1_2"},
        {"0_1_2", "0_2_2"},
        {"0_0_1", "0_2_2"},
    ]
    assert conflict_set == unordered(expected)



import networkx as nx
from motile import TrackGraph


def graph_to_nx(graph: TrackGraph) -> nx.DiGraph:
    """Convert a motile TrackGraph into a networkx DiGraph.

    Args:
        graph (TrackGraph): TrackGraph to be converted to networkx

    Returns:
        nx.DiGraph: Directed networkx graph with same nodes, edges, and attributes.
    """
    nx_graph = nx.DiGraph()
    nodes_list = list(graph.nodes.items())
    nx_graph.add_nodes_from(nodes_list)
    edges_list = [
        (edge_id[0], edge_id[1], data) for edge_id, data in graph.edges.items()
    ]
    nx_graph.add_edges_from(edges_list)
    return nx_graph



from collections import Counter

import numpy as np
import pytest
from motile_toolbox.candidate_graph import EdgeAttr, get_candidate_graph
from motile_toolbox.candidate_graph.compute_graph import (
    get_candidate_graph_from_points_list,
)


def test_graph_from_segmentation_2d(segmentation_2d, graph_2d):
    # test with 2D segmentation
    cand_graph, _ = get_candidate_graph(
        segmentation=segmentation_2d,
        max_edge_distance=100,
        iou=True,
    )
    assert Counter(list(cand_graph.nodes)) == Counter(list(graph_2d.nodes))
    assert Counter(list(cand_graph.edges)) == Counter(list(graph_2d.edges))
    for node in cand_graph.nodes:
        assert Counter(cand_graph.nodes[node]) == Counter(graph_2d.nodes[node])
    for edge in cand_graph.edges:
        print(cand_graph.edges[edge])
        assert (
            pytest.approx(cand_graph.edges[edge][EdgeAttr.IOU.value], abs=0.01)
            == graph_2d.edges[edge][EdgeAttr.IOU.value]
        )

    # lower edge distance
    cand_graph, _ = get_candidate_graph(
        segmentation=segmentation_2d,
        max_edge_distance=15,
    )
    assert Counter(list(cand_graph.nodes)) == Counter(["0_1", "1_1", "1_2"])
    assert Counter(list(cand_graph.edges)) == Counter([("0_1", "1_2")])


def test_graph_from_segmentation_3d(segmentation_3d, graph_3d):
    # test with 3D segmentation
    cand_graph, _ = get_candidate_graph(
        segmentation=segmentation_3d,
        max_edge_distance=100,
    )
    assert Counter(list(cand_graph.nodes)) == Counter(list(graph_3d.nodes))
    assert Counter(list(cand_graph.edges)) == Counter(list(graph_3d.edges))
    for node in cand_graph.nodes:
        assert Counter(cand_graph.nodes[node]) == Counter(graph_3d.nodes[node])
    for edge in cand_graph.edges:
        assert pytest.approx(cand_graph.edges[edge], abs=0.01) == graph_3d.edges[edge]


def test_graph_from_multi_segmentation_2d(
    multi_hypothesis_segmentation_2d, multi_hypothesis_graph_2d
):
    # test with 2D segmentation
    cand_graph, conflict_set = get_candidate_graph(
        segmentation=multi_hypothesis_segmentation_2d,
        max_edge_distance=100,
        iou=True,
    )
    assert Counter(list(cand_graph.nodes)) == Counter(
        list(multi_hypothesis_graph_2d.nodes)
    )
    assert Counter(list(cand_graph.edges)) == Counter(
        list(multi_hypothesis_graph_2d.edges)
    )
    for node in cand_graph.nodes:
        assert Counter(cand_graph.nodes[node]) == Counter(
            multi_hypothesis_graph_2d.nodes[node]
        )
    for edge in cand_graph.edges:
        assert (
            pytest.approx(cand_graph.edges[edge][EdgeAttr.IOU.value], abs=0.01)
            == multi_hypothesis_graph_2d.edges[edge][EdgeAttr.IOU.value]
        )
    # TODO: Test conflict set

    # lower edge distance
    cand_graph, _ = get_candidate_graph(
        segmentation=multi_hypothesis_segmentation_2d,
        max_edge_distance=14,
    )
    assert Counter(list(cand_graph.nodes)) == Counter(
        list(multi_hypothesis_graph_2d.nodes)
    )
    assert Counter(list(cand_graph.edges)) == Counter(
        [("0_0_1", "1_0_2"), ("0_0_1", "1_1_2"), ("0_1_1", "1_1_2")]
    )


def test_graph_from_points_list():
    points_list = np.array(
        [
            [0, 1, 1, 1],
            [2, 3, 3, 3],
            [1, 2, 2, 2],
            [2, 6, 6, 6],
            [2, 1, 1, 1],
        ]
    )
    cand_graph = get_candidate_graph_from_points_list(points_list, max_edge_distance=3)
    assert cand_graph.number_of_edges() == 3
    assert len(cand_graph.in_edges(3)) == 0



from enum import Enum


class NodeAttr(Enum):
    """Node attributes that can be added to candidate graph using the toolbox.
    Note: Motile can flexibly support any custom attributes. The toolbox provides
    implementations of commonly used ones, listed here.
    """

    POS = "pos"
    TIME = "time"
    SEG_ID = "seg_id"
    SEG_HYPO = "seg_hypo"


class EdgeAttr(Enum):
    """Edge attributes that can be added to candidate graph using the toolbox.
    Note: Motile can flexibly support any custom attributes. The toolbox provides
    implementations of commonly used ones, listed here.
    """

    IOU = "iou"
